<?php
// +----------------------------------------------------------------------+
// | This source file is subject to version 2.0 of the PHP license,       |
// | that is bundled with this package in the file LICENSE, and is        |
// | available at through the world-wide-web at                           |
// | http://www.php.net/license/2_02.txt.                                 |
// | If you did not receive a copy of the PHP license and are unable to   |
// | obtain it through the world-wide-web, please send a note to          |
// | license@php.net so we can mail you a copy immediately.               |
// +----------------------------------------------------------------------+
// | Authors: Jason Rust <jrust@rustyparts.com>                           |
// +----------------------------------------------------------------------+

// {{{ includes

require_once 'PEAR.php';

// }}}

/**
 * A class to convert a local html file to a pdf file on the fly.
 * Relies on some external programs to make things tick,
 * all of which only run on Linux/Unix to my knowledge
 *
 * These include:
 * html2ps - Does the initial conversion of html to a postscript file.
 *           more information about it here: http://www.tdb.uu.se/~jan/html2ps.html
 *           You will especially want to read the User's Guide for html2ps 
 *           if this script is not making the pdf look like you want, since html2ps is the
 *           heart of what converts the html page.
 * ps2pdf -  This comes with the Ghostscript package, and can be found here:
 *           http://www.cs.wisc.edu/~ghost/  This package is normally installed as an
 *           RPM on RedHat systems, so if you're using that OS you shouldn't have to worry.
 * curl   -  Or some program that grabs documents off the web (lynx, w3m, etc.).  HOWEVER, for whatever 
 *           reason only curl has worked in my tests when this script is running as a web script and not 
 *           just a script from the command line
 *
 * @author Jason Rust <jrust@rustyparts.com>
 * @version 2.0
 * @copyright The PHP License
 */
class HTML_ToPdf{
    // {{{ properties

    /**
     * The full path to the file we are parsing
     * @var string
     */
    var $htmlFile = '';

    /**
     * The full path to the output file 
     * @var string
     */
    var $pdfFile = '';

    /**
     * The default domain for relative images 
     * @var string
     */
    var $defaultDomain = '';

    /**
     * The path to the html2ps executable 
     * @var string
     */
    var $html2psPath = '/usr/bin/html2ps';

    /**
     * The path to the ps2pdf executable 
     * @var string
     */
    var $ps2pdfPath = '/usr/bin/ps2pdf';

    /**
     * The path to your get URL program, including options to get headers
     * @var string
     */
    var $getUrlPath = '/usr/bin/curl -i';

    /**
     * Whether or not to try and parse the CSS in the html file and use it in creating the pdf
     * @var bool
     */
    var $useCSS = true;

    /**
     * Other styles to use when parsing the page
     * @var string
     */
    var $additionalCSS = '';

    /**
     * Show the page in color?
     * @var bool
     */
    var $pageInColor = true;

    /**
     * Show the images be in grayscale?
     * @var bool
     */
    var $grayScale = false;

    /**
     * Scale factore for the page 
     * @var int 
     */
    var $scaleFactor = 1;

    /**
     * Whether to underline links or not 
     * @var bool 
     */
    var $underlineLinks = null;

    /**
     * Default html2ps configuration that we use (is parsed before being used, though)
     * @var string
     */
    var $html2psrc = '
        header {               /* Running page header */
          odd-left: $T;        /* Document title */
          odd-right: $H;       /* Current main heading */
          even-left: $H;       /* Ditto */
          even-right: "$[author]"; /* Document author (from <META NAME="Author"...>) */
        }
        footer {               /* Running page footer */
          center: "- $N -";    /* Page number */
        }
        option {
          titlepage: 0;         /* do not generate a title page */
          colour: %pageInColor%; /* create the page in color */
          underline: %underlineLinks%;         /* underline links */
          grayscale: %grayScale%; /* Make images grayscale? */
          scaledoc: %scaleFactor%; /* Scale the document */
        }
        package {
          geturl: %getUrlPath%; /* path to the geturl */
        }
        showurl: 0;             /* do not show the url next to links */';

    /**
     * We use this to store the html file to a string for manipulation 
     * @var string
     */
    var $_htmlString = '';

    // }}}
    // {{{ constructor

    /**
     * Initializes the class
     *
     * @param string $in_htmlFile The full path to the html file to convert
     * @param string $in_domain The default domain name for images that have a relative path
     * @param string $in_pdfFile (optional) The full path to the pdf file to output.  
     *               If not given then we create a temporary name.
     *
     * @access public
     * @return void
     */
    function HTML_ToPdf($in_htmlFile, $in_domain, $in_pdfFile = null)
    {
        $this->htmlFile = $in_htmlFile;
        $this->defaultDomain = $in_domain;
        
        if (is_null($in_pdfFile)) {
            $this->pdfFile = tempnam ('', 'PDF-');
        }
        else {
            $this->pdfFile = $in_pdfFile;
        }
    }

    // }}}
    // {{{ addHtml2PsSettings()

    /**
     * Adds on more html2ps settings to the end of the default set of settings 
     *
     * @param string $in_settings The additional settings 
     *
     * @access public
     * @return void
     */
    function addHtml2PsSettings($in_settings) {
        $this->html2psrc .= "\n" . $in_settings;
    }

    // }}}
    // {{{ setUseColor()

    /**
     * Set whether to use color or not when creating the page
     *
     * @param bool $in_useColor Use color?
     *
     * @access public
     * @return void
     */
    function setUseColor($in_useColor) {
        $this->pageInColor = $in_useColor;
    }

    // }}}
    // {{{ setUseCSS()

    /**
     * Set whether to try and use the CSS in the html page when creating
     * the pdf file
     *
     * @param bool $in_useCSS Use CSS found in html file?
     *
     * @access public
     * @return void
     */
    function setUseCSS($in_useCSS) {
        $this->useCSS = $in_useCSS;
    }

    // }}}
    // {{{ setAdditionalCSS()

    /**
     * Set additional CSS to use when parsing the html file 
     *
     * @param string $in_css The additional css
     *
     * @access public
     * @return void
     */
    function setAdditionalCSS($in_css) {
        $this->additionalCSS = $in_css;
    }

    // }}}
    // {{{ setGetUrl()

    /**
     * Sets the get url which is used for retrieving images from the html file
     * needs to be the full path to the file with options to retrive the headers
     * as well.
     *
     * @param string $in_getUrl The get url program 
     *
     * @access public
     * @return void
     */
    function setGetUrl($in_getUrl) {
        $this->getUrl = $in_getUrl;
    }

    // }}}
    // {{{ setGrayScale()

    /**
     * Sets the gray scale option for images 
     *
     * @param bool $in_grayScale Images should be in grayscale? 
     *
     * @access public
     * @return void
     */
    function setGrayScale($in_grayScale) {
        $this->grayscale = $in_grayScale;
    }

    // }}}
    // {{{ setUnderlineLinks()

    /**
     * Sets the option to underline links or not 
     *
     * @param bool $in_underline Links should be underlined? 
     *
     * @access public
     * @return void
     */
    function setUnderlineLinks($in_underline) {
        $this->underlineLinks = $in_underline;
    }

    // }}}
    // {{{ setScaleFactor()

    /**
     * Sets the scale factor for the page.  Less than one makes it smaller,
     * greater than one enlarges it.
     *
     * @param int $in_scale Scale factor 
     *
     * @access public
     * @return void
     */
    function setScaleFactor($in_scale) {
        $this->scaleFactor = $in_scale;
    }

    // }}}
    // {{{ setHtml2Ps()

    /**
     * Sets the path to the html2ps program 
     *
     * @param string $in_html2ps The html2ps program 
     *
     * @access public
     * @return void
     */
    function setHtml2Ps($in_html2ps) {
        $this->html2psPath = $in_html2ps;
    }

    // }}}
    // {{{ setPs2Pdf()

    /**
     * Sets the path to the ps2pdf program 
     *
     * @param string $in_ps2pdf The ps2pdf program 
     *
     * @access public
     * @return void
     */
    function setPs2Pdf($in_ps2pdf) {
        $this->ps2pdfPath = $in_ps2pdf;
    }

    // }}}
    // {{{ convert()

    /**
     * Convert the html file into a pdf file
     *
     * @access public
     * @return string The path to the pdf file 
     */
    function convert()
    {
        // make sure html file exists
        if (!file_exists($this->htmlFile) && !preg_match(':^(f|ht)tp\://:i', $this->htmlFile)) {
            return new HTML_ToPDFException("Error: The HTML file does not exist: $this->htmlFile");
        }

        // first make sure we can execute the programs
        if (!@is_executable($this->html2psPath)) {
            return new HTML_ToPDFException("Error: html2ps [$this->html2psPath] not executable");
        }

        if (!@is_executable($this->ps2pdfPath)) {
            return new HTML_ToPDFException("Error: ps2df [$this->ps2pdfPath] not executable");
        }

        // this can take a while with large files
        set_time_limit(60);

        // read the html file in so we can modify it
        $this->_htmlString = @implode('', @file($this->htmlFile));
        // grab extra CSS
        $this->additionalCSS .= $this->_getCSSFromFile();
        // modify the conf file
        $this->_modifyConfFile();

        // try and replace relative images with the default domain
        $this->_htmlString = preg_replace(':<img (.*?)src=["\']((?!http\://).*?)["\']:i', '<img \\1 src="http://'.$this->defaultDomain.'/\\2"', $this->_htmlString);
        // html2ps messes up on several form elements
        $this->_htmlString = preg_replace(':<input (.*?)type=["\']?(hidden|submit|button|image|reset|file)["\']?.*?>:i', '<input />', $this->_htmlString);

        // the conf file has to be an actual file
        $configFile = tempnam('', 'CONF-');
        $fp = fopen($configFile, 'w');
        fwrite($fp, $this->html2psrc);
        fclose($fp);

        // make the temporary html file 
        $htmlFile = tempnam('', 'HTML-');
        $fp = fopen($htmlFile, 'w');
        fwrite($fp, $this->_htmlString);
        fclose($fp);

        // needa temporary postscript file as well
        $psFile = tempnam('', 'PS-');

        exec($this->html2psPath . ' -f ' . $configFile . ' -o ' . $psFile . ' \'' . escapeshellcmd($htmlFile) .  '\'');
        exec($this->ps2pdfPath . ' ' . $psFile . ' \'' . escapeshellcmd($this->pdfFile) .  '\'');

        unlink($htmlFile);
        unlink($configFile);
        unlink($psFile);

        return $this->pdfFile;
    }

    // }}}
    // {{{ _modifyConfFile ()

    /**
     * Modify the config file and put in our custom variables
     *
     * @access private
     * @return void
     */

    function _modifyConfFile()
    {
        // first determine if we should try and figure out underline link option, based on css
        if (is_null($this->underlineLinks)) {
            if (preg_match(':a\:link {.*?text-decoration\: (.*?);.*?}:is', $this->additionalCSS, $matches) &&
                is_int(strpos($matches[1], 'none'))) {
                $this->underlineLinks = false;
            }
            else {
                $this->underlineLinks = true;
            }
        }

        $this->html2psrc = str_replace('%scaleFactor%', $this->scaleFactor, $this->html2psrc);
        $this->html2psrc = str_replace('%getUrlPath%', $this->getUrlPath, $this->html2psrc);
        // we convert booleans into numbers
        $this->html2psrc = str_replace('%pageInColor%', (int) $this->pageInColor, $this->html2psrc);
        $this->html2psrc = str_replace('%grayScale%', (int) $this->getUrlPath, $this->html2psrc);
        $this->html2psrc = str_replace('%underlineLinks%', (int) $this->underlineLinks, $this->html2psrc);

        // add the global container
        $this->html2psrc = '
        @html2ps {
          '.$this->html2psrc.'
        }
        /* throw on these to make the page look nice */
        @page:left {
          margin-left: 2cm;
          margin-right: 3cm;
        }
        @page:right {
          margin-left: 3cm;
          margin-right: 2cm;
        }
        '.$this->additionalCSS;
    }

    // }}}
    // {{{ _getCSSFromFile()

    /**
     * Try to get the CSS from the html file and use it in creating the
     * PDF file.  If we find CSS we'll add it to the CSS string
     *
     * @access private
     * @return string Any CSS found 
     */
    function _getCSSFromFile()
    {
        if ($this->useCSS) {
            $cssFound = '';
            // first try to find inline styles
            if (preg_match(':<style.*?>(.*?)</style>:is', $this->_htmlString, $matches)) {
                $cssFound = $matches[1];
                // replace it with nothing in the html since it messes up html2ps
                $this->_htmlString = preg_replace(':<style.*?>.*?</style>:is', '', $this->_htmlString);
            }
            elseif (preg_match(':<link .*? href=["\'](.*?)["\'].*?text/css.*?>:i', $this->_htmlString, $matches)) {
                $cssFound = preg_replace(':(^(?!http\://).*):i', 'http://'.$this->defaultDomain.'/\\1', $matches[1]);
                $cssFound = implode('', file($cssFound));
            }

            // only takes a:link attribute
            $cssFound = preg_replace(':a +{:i', 'a:link {', $cssFound);

            // font-size: word causes a crash
            $cssFound = preg_replace(':font-size\: *(\w*);:ie', '$this->_convertFontSize("\\1")', $cssFound);

            return $cssFound;
        }
        else {
            return '';
        }
    }

    // }}}
    // {{{ _convertFontSize()

    /**
     * Converts textual font size to a numberic representation.
     *
     * @param string $in_fontString The font size specification
     *
     * @access private
     * @return string The font-size attribute with size in pt 
     */
    function _convertFontSize($in_fontString)
    {
        switch ($in_fontString) {
            case 'xx-small':
                $size = 6; 
                break;
            case 'x-small':
                $size = 8;
                break;
            case 'small':
                $size = 10;
                break;
            case 'medium':
                $size = 12;
                break;
            case 'large':
                $size = 14;
                break;
            case 'x-large':
                $size = 16;
                break;
            case 'xx-large':
                $size = 18;
                break;
            default:
                $size = 12;
                break;
        }

        return 'font-size: ' . $size . 'pt;';
    }

    // }}}
}

// {{{ HTML_ToPDFException

class HTML_ToPDFException extends PEAR_Error {
    var $classname             = 'HTML_ToPDF';
    var $error_message_prepend = 'Error: ';

    function HTML_ToPDFException ($message)
    {
        $this->PEAR_Error($message);
    }
}

// }}}
